;Source File     : OkKey35.SC
;Submitted By    : Alan Zenreich, PAL By Example.
;Version         : 1.5
;Last revision   : 1/10/90 
;                  Updated to work with Paradox 3.5, 3.0 and 3.01.
;Description     : These procedures trap for key violations in "real-time"
;                  during the CoEdit of a table.
;                  See text below for full commentary
;                  Also a routine for an intelligent COPYFROMARRAY that works
;                  with detail tables.  Finally, there are message
;                  procs and error procs to support the key violations procs.
;
;
; These procedures are excerpted from the utility scripts of WaitPlus Pro,
; our data control program for Paradox.  Some of them were also published
; in our book:
;    The Paradox Programmer's Guide:  PAL By Example
;
; You might ask "Why are Alan and Jim sharing all of these procedures?"
; We're publishing them because key violation checking is very tricky in the
; current Paradox versions, but very important to implement.
; Unique keys is an integral part of any well designed system, and these
; procedures can help maintain the integrity of your data.
;
; The code below uses many techniques to get the desired effect, and is
; optimized to take advantage of version specific features of Paradox.
; The issues are complicated, but the source code is pretty well commented.
;
; This is only a small sampling of the source code provided in WaitPlus Pro.
; You are granted rights to use this code in your applications provided that
; somewhere in your code you state:
;   "Portions of this program is copyright 1991 PAL By Example"
; You must also agree not to use this code in a product that generally
; competes with WaitPlus Pro.  This agreement is simliar in intent to
; the Paradox runtime agreement.
;
; You might be interested in the products/programs that we market:
;   The Paradox Programmer's Guide:  PAL By Example
;   ParaLex, the table documentor
;   PlayRight Pro, the Paradox script editor
;   TxtView, memo fields for Paradox
;   WaitPlus Pro, data control for Paradox
;
;  If you would like additional information contact
;    PAL By Example
;    One Newark Street
;    Hoboken, NJ  07030
;    Phone (201) 795-5144
;    FAX   (201) 795-9237
;
;  We hope you find this code as useful as we do.
;
;  Enjoy!
;
;  Alan Zenreich and James Kocis.
;
;  Contents:
;---------------------------------------------------------------------
;  Beeper.u(delay.n)                    ; Beeps until key is pressed.
;  TopMsg.u(line1.a, line2.a, delay.v)  ; Centered two line message
;                                       ; with optional beeps.
;  OkKey35.l(requiredcheck.l)           ; Ok Key procedure that is sensitive
;                                       ; to features in Paradox 3.5.  It is
;                                       ; typically called by OkKey3.l()
;  OkKey3.l(requiredcheck.l)            ; Does current record pose a key
;                                       ; conflict? Calls OkKey35 if appropriate.
;  CopyFromArray3.u(arrayname.a,        ; Copies array to detail starting
;                   requiredcheck.l)    ; with the first field placed on
;                                       ; the form. Substitute for
;                                       ; COPYFROMARRAY.
;  InvalidArrayRecover.u(arrayname.a)   ; Special error procedure to handle
;                                       ; COPYFROMARRAY errors when array
;                                       ; contains invalid data.
;  OKKey2.l()                           ; Key Violation proc for unlinked
;                                       ; tables.  See notes in OkKey3.l()
;  OopsPrint.u()                        ; Error handling procedure.

;---------------------------------------------------------------------
; Beeper.u()
; Beeper.u clears the keyboard buffer and beeps at the user until any
; key is pressed.  The delay between beeps may vary between 1 and
; 30000 milliseconds.
;
;  Example:  Beeper.u(1000)
;            Sounds a beep every second until a key is pressed.

PROC Beeper.u(delay.n)
  ; delay.n   ; Number of milliseconds between beeps, 1 - 30000
  PRIVATE n,                    ; Any characters in the buffer.
          proc.a                ; Name of procedure
  proc.a = "Beeper.u"
  WHILE CHARWAITING()           ; User might be typing ahead of the program.
    n = GETCHAR()               ; Grab keystroke from buffer assign to n.
  ENDWHILE
  IF delay.n > 30000 THEN
    delay.n = 1000              ; If more than 30000, set delay.n to one second.
  ENDIF
  WHILE NOT CHARWAITING()       ; Waits for keystroke from operator.
    BEEP BEEP                   ; Two quick beeps.
    SLEEP delay.n               ; Pauses for the amount specified by delay.n.
  ENDWHILE
  n = GETCHAR()                 ; Retrieves the operator keystroke.
ENDPROC
WRITELIB libname.a Beeper.u
RELEASE PROCS Beeper.u

;---------------------------------------------------------------------
; TopMsg.u()
; Procedure to place two lines of text centered at the top of screen.
; The user can easily get accustomed to looking to the top of screen for all
; instructions, errormessages and menus.  Messages never obscure data.
; This proc uses the one above, Beeper.u(), to sound a tone at regular
; intervals. Varying the delay.n parameter allows you to adjust the
; "urgency" of the message.
;
; Example:  TopMsg.u("This is an error", "Press any key...",1000)

PROC TopMsg.u(line1.a, line2.a, delay.v)
  ; line1    ; Text to display at the top of the screen'
  ; line2    ; Text to display on the second line of the screen
             ; Three shortcuts have been implemented, when evaluating line2.a
             ; If line2.a = "" second line will read "Press any key".
             ; If "X", second line will read "Press any key to exit".
             ; If "C", second line will read "Press any key to continue".
             ; So if you want a blank second line make line2.a = "  ".
  ; delay.v  ; Amount of delay between beeps.
             ; If 0, it beeps once and waits for a key.
             ; If a number between 1 and 30000, beeps until a key is pressed,
             ; Using the delay.v as the pause between beeps.
             ; If "", it simply puts up the message, relying on the calling
             ; program to get keystrokes and clean up display.

  PRIVATE
  n,       ; Used to get keyboard character
  pak.a,   ; String for "Press any key"
  proc.a   ; name of procedure

  proc.a = "TopMsg.u"

  IF VERSION() <3 THEN
    STYLE REVERSE                    ; Use inverse video.
  ELSE
    STYLE ATTRIBUTE SYSCOLOR(3)      ; Use same color as Paradox messages.
  ENDIF
  @ 0,0 ?? FORMAT("W80,AC",line1.a)  ; Center the first line.
  IF LEN(line2.a) <= 1 THEN
    pak.a = "Press any key"
    SWITCH
      CASE line2.a = "":
        line2.a = pak.a
      CASE UPPER(line2.a) = "X":
        line2.a = pak.a + " to exit"
      CASE UPPER(line2.a) = "C":
        line2.a = pak.a + " to continue"
    ENDSWITCH
  ENDIF
  ?? FORMAT("W80,AC", line2.a)     ; Center the second line.
  STYLE
  SWITCH
    CASE delay.v = "":             ; Return to calling program.
      RETURN
    CASE delay.v = 0 :
      BEEP BEEP                    ; Beep twice and get a character.
      n = GETCHAR()
    OTHERWISE :                    ; Most often used, this beeps at the user
      Beeper.u(delay.v)            ; until a key is pressed.
  ENDSWITCH
  @1,0 CLEAR EOL                   ; Clean up after messages, except for
  @0,0 CLEAR EOL                   ; delay.v = "".
ENDPROC
WRITELIB libname.a TopMsg.u
RELEASE PROCS      TopMsg.u
;-------------------------------------------------------------------------
;
; These following procedures were developed to trap for key violations
; in Coedit mode in real time.

; First a little background on why this is necessary. In interactive
; Paradox in CoEdit mode, a user is made aware of key violations with
; a message:
;
;"Key exists -- press [Alt] [L] to confirm [Alt] [K] to see existing record"

; A key conflict may only occur when trying to post the record. This
; can happen either when the user adds a new record or edits an existing
; record. Key violation checking under script control in real time is a
; somewhat more complicated matter. Here are some considerations:
;
; Newly posted records in a keyed table "fly away" from the cursor to
; their proper location in the table. Unless you are at the proper
; location when adding the record (or changing its key), the record
; just changed appears to disappear. Better would be to to remain on
; the record, and not let it fly away (to "travel" with it).
;
; Other situations:
;
;    A record is posted with a valid key and it is edited and
;    replaced with a conflicting key.
;
;    A record, not yet posted to table (a new record) should
;    stay unposted until you are ready to commit the changes.
;
;    If editing an existing record changes the key, how does
;    this affect other records that depend on it (should the user
;    be allowed to change the key in the first place?)
;
;    If it okay to change a key and the user accidentally changes
;    it to an existing key, then we have to deal with it immediately,
;    Do you want to show the user the offending key, and have them
;    toggle? Do you want to tell the user that the key is already on
;    file, and reject the change?
;
;    What happens if the user wants to "UNDO" his or her changes?
;    When does the record disappear from other nodes on the network?
;    If another user attempts to edit the record on their node, do
;    they get a "Record deleted by another user" message?
;
; Obviously we would prefer new PAL functionality (perhaps a new
; function, ISKEYVALID() and a DETAILFROMARRAY command).

; There are three key violation checking procedures here.
; OkKey35.l() uses features of Paradox 3.5 where available.
; It is the fastest version, and is the safest one when confronted by
; invalid data (possible when editing an existing record where the data no
; longer passes the validity checks).  The other two procedures need to
; attempt to delete the record and reinsert it.  At that point the key
; violation can be evaluated.  If you study the 3.5 technique below, and
; compare it to the older techniques, you'll see some of the issues raised.


; When you use the OkKey2.l() or OkKey3.l() procedures, if a key conflict
; exists, the record is deleted and then reinserted to make it an unposted
; record.
; OkKey3.l(requiredcheck.l) is used to return a TRUE or FALSE
; to inform calling program whether or not the proposed key is "OK".
; It is useful just about any time you need to see if there is a
; duplicate key (for example, in a DEToolkit Good Depart field
; procedure).
;
; CopyFromArray3.u(arrayname.a, requiredcheck.l) is another necessary
; evil. It simulates a COPYFROMARRAY but will work in an embedded
; linked detail table in a multitable form, where the linked fields
; cannot be placed explicitly.  It also takes care of some less obvious
; problems.  New to version 1.5 is error trapping to deal with invalid data
; found in the array.  The errorchecking is extensive and commented in the
; code.
;
; At this writing everything appears to work, as it has undergone
; extensive testing under a variety of LinkTypes.
;---------------------------------------------------------------------------


; Purpose: Checks for key violation in coedit mode for Paradox 3.5.
; Note: It is OkKey35 because it uses Paradox 3.5 commands.
;
; The proc returns a TRUE if the key is okay, and a FALSE if not.
; Because we can determine the RECORDSTATUS(), the record's status is not
; changed (unlike the OkKey3.l proc that deletes and reinserts a record),
; except for the fact that the record is locked at the end of the key violation
; checking.  The procedure also goes to great lengths to avoid conflicts
; with existing invalid data.
;
; Paradox 3.5 specific features are coded with EXECUTE statements so they
; can be compiled into a 3.0 library yet still take advantage of the
; 3.5 feature set (very useful in a network with mixed Paradox versions)
;
;Examples l = OkKey35.l(TRUE)   ; Make sure REQUIREDCHECK is turned back on.
;         l = OkKey35.l(FALSE)  ; REQUIREDCHECK will remain off.
;         l = OkKey35.l("")     ; REQUIREDCHECK will be turned back ON.
;

PROC OkKey35.l(requiredcheck.l)
  ; requiredcheck.l  ; Determines if required fields should be
  ;                  ; reinstated when leaving the proc.
                     ; Should be TRUE if you want to make sure that
                     ; REQUIREDCHECK ON is in effect when exiting
                     ; Copyfromarray3.u, which is called by this proc.
                     ; If requiredcheck.l is any value other than FALSE,
                     ; REQUIREDCHECK will be set back to the normal
                     ; ON status.
  ; RETURNS TRUE if key is okay, FALSE if key already exists

  PRIVATE
  field.a,         ; Current field name.
  r,               ; Temporary array.
  a,               ; Window() message.
  ok.l,            ; Is record okay (no key conflict)?
  error.n,         ; Errorcode for keyviolation or write lock.
  keys.n,          ; number of keyfields.
  maxkeys.n,       ; maximum number of keys to check for.
  fields.n,        ; number of fields.
  newkey.l,        ; has the key changed.
  values.a,        ; Used to locate record in the event that it flies away.
  n,               ; Counter
  old.r,           ; Values in unedited record, used for comparison.
  proc.a           ; Name of procedure

  proc.a = "OkKey35.l"

      ; The next few lines allow you to specify an alternate okkey procedure,
      ; perhaps necessary under special circumstances, for example when an
      ; embedded table is also explicitly place on the workspace. To invoke
      ; an alternate key violation procedure, assign a variable called
      ; okkeyproc.a to the name of the alternate procedure prior to invoking
      ; the okkey3.l procedure. The alternate procedure must return a TRUE
      ; if the key is okay, or FALSE if there is a keyviolation.
      ;
      ; This can be handy for performance gains by making the key checking
      ; more specific to the particular needs of the table. Remember
      ; to release or reassign okkeyproc.a = "" when no longer needed.

  IF IsAssigned (okkeyproc.a) AND okkeyproc.a <> "" THEN
    EXECPROC okkeyproc.a    ; Do the special procedure and return its value.
    RETURN RETVAL
  ENDIF
      ; If no special keyfield proc has been assigned, we continue from this point
  keys.n = NKeyFields(TABLE())
  IF keys.n = 0 OR linktype()="1-1" THEN
    ok.l = true      ; Cannot possibly be a key conflict.
  ELSE               ; There must be key fields.
            ; Now on to the keyviolation checking...
    RequiredCheck OFF     ; Helps deal with required valcheck and empty fields.
    ImageRights           ; Make sure that we have rights to do this.
    field.a = FIELD()   ; Just in case RESYNCKEY manages to jump you to

    IF IsFormView() THEN  ; We need to see if this is a master form
      RESYNCKEY  ; Shows a key conflict without unlocking-- records won't fly away.
      window.a = WINDOW()
      IF window.a = "Only relevant when editing master table in a multi-table form" THEN
        master.l = FALSE     ; This is not a master form
      ELSE
        master.l = TRUE
        error.n = ERRORCODE()    ; Get the errorcode
        ok.l = error.n <> 53 AND error.n <> 3
        ; Error.n 53 is set if key exists, error.n 3 if a Write Lock is on table
      ENDIF
    ELSE
      master.l =false   ; If not in form view, we can't possibly be in a master
    ENDIF
    IF NOT master.l THEN  ; lets see if this is a new unposted record
      EXECUTE "new.l = RECORDSTATUS(\"New\")"
      EXECUTE "modified.l = RECORDSTATUS(\"Modified\")"
      SWITCH
        CASE new.l :  ; A new unposted record.
          KeyLookup   ; Attempt toggle to possible offending key.
          error.n = ERRORCODE()
          ok.l = error.n = 56  ; If it can't find lookup, then key is okay.
          IF ok.l THEN         ; Lookupkey not found or another user locked
                               ; the offending key (which means it exists).
            LockRecord         ; Record posts, and cursor remains with it.
          ELSE
            KeyLookup          ; Return to record values as operator entered.
          ENDIF
        CASE (NOT new.l) AND (NOT modified.l):
          ; The record exists, and has not been modified, therefore it must be ok
          ok.l = true
        OTHERWISE:
          ; The record is existing and modified (therefore obviously locked.)
          ; We will unlock the record and try to locate it again after it "flies away"
          ; Granted, there is a remote possiblity that another user can lock the record
          ; or even delete the record before we get to it, but it probably has about
          ; the same possibility as someone sucessfully posting a conflicting record
          ; between the "lookupkey" and lockrecord statements in the new record routine
          ; posted above.  We'll check for the possible multi-user conflict below.

          ; We are going to try to UNLOCK (which posts) the record and then
          ; re-lock it.
          ; When the record is unlocked, we need to be able to find it again.
          ; First we'll copy the values to an array.
          COPYTOARRAY r      ; Save the edited state of the record.
          ; Now we'll construct a string to execute that will find the record in
          ; the event that it flies away.
          ; For example, let's assume there are 3 keys.  To find the records
          ; we'd have to issue LOCATE r[2],r[3],r[4] which will find the
          ; unique record.  Remember that the first element of the array is the
          ; table name so it is not used.
          ; The second element is the first key field.

          ; Now let's construct the search string to be EXECUTE'd
          values.a=""  ; initialize string
          ; Now let's make sure that we don't have more than 20 keys
          ; because the EXECUTE string below would be too long.
          maxkeys.n = MIN(keys.n, 21)
          FOR n FROM 1 TO maxkeys.n
            IF n >1 THEN
              values.a = values.a +","     ; add a comma before the next value
            ENDIF
            values.a = values.a + "r["+STRVAL(n+1)+"]"
          ENDFOR
          ; We now have a string that might look like
          ; "r[2],r[3],r[4]"

          ; Now we have to check to see if there is only one key.  If so,
          ; we'll have to move to the keyfield because LOCATE r[2] would only
          ; locate the current column.  If we have more fields, we can append
          ; another value
          IF keys.n = 1 AND ARRAYSIZE(r)>2 THEN  ; there are more fields
            values.a = values.a+",r[3]"
            ; so we can locate regardless of current field
          ENDIF
           ; Now, the logic that follows is
           ; 1) If the non key fields have been modified and the key fields haven't
           ;    been touched, KEYLOOKUP will never show a conflicting key.
           ; 2) If the key field has been touched, KEYLOOKUP will attempt to
           ;    find existing key.
          fields.n=NFields(TABLE())  ; Get the number of fields in the table

           ; Now we have to determine if the keys have changed.
           ; (The record is at this point unquestionably locked)
          Undo    ; Return the record to an unchanged, unlocked state.
          LockRecord  ; Let's not lose control of the record
          CopyToArray old.r  ; The unedited version of the record
          newkey.l = FALSE   ; Let's assume that the key has not been changed
          changed.l= FALSE   ; Let's assume that the record has not been changed
          FOR n FROM 2 TO fields.n + 1
            IF r[n] <> old.r[n] THEN
              changed.l = true
              ; There has been a change.  Let's see if a key has been changed
              IF n <= keys.n + 1 THEN
                newkey.l=true
              ENDIF
              QUITLOOP
            ENDIF
          ENDFOR
          ; If any fields have been changed, we have to replace any changed values that we've undone
          IF changed.l THEN
            Menu {Image} {Zoom} {Field}
            FOR n FROM 1 TO fields.n
              IF r[n+1] <> old.r[n+1] THEN
                EXECUTE "[" + MENUCHOICE() + "]= r[n+1]"
              ENDIF
              Right
            ENDFOR
            ; For example, if the first field is [Cust#] the execute would prepare:
            ; [Cust#] = r[2]      ; The second element of the array.
            CtrlBreak             ; Back out of the zoom/field.
            ; Now we have the original record restored
          ENDIF
          ok.l = true   ;Let's assume that the key is ok

          IF newkey.l THEN   ; We have to try and see if there is a conflicting key
            KeyLookup        ; Attempt toggle to possible offending key.
            error.n = ERRORCODE()
            ok.l = error.n = 56  ; If it can't find a conflicting lookup,
            IF not ok.l THEN     ; Lookupkey found return to record values
              KeyLookup          ; as operator entered them.
            ELSE                 ; The lookup key has not been found, or another
                                 ; use has locked the offending key (which means it exists).
              UnLockRecord       ; Record posts, and can fly away.
              IF ARRAYSIZE(r) = 2 THEN  ; We must have only one field
                ; Make sure we move to the field for LOCATE, whether in table or form view
                CtrlEnd
              ENDIF
              EXECUTE "LOCATE " + values.a
              IF NOT RETVAL THEN
                TopMsg.u("The modified record posted correctly but cannot be located now",
                         "Your cursor may wind up on an incorrect record",1000)
              ELSE   ; we found the record
                LockRecord  ; let's regain control over the record
                IF NOT RETVAL AND ERRORCODE()<>55 THEN
                  ; Another user grabbed the record
                  TopMsg.u("The modified record posted correctly but cannot be locked now",
                           "You no longer have control over the record",1000)
                ENDIF; If we couldn't lock record
              ENDIF  ; If we located record
            ENDIF    ; If key is ok
          ENDIF      ; If it is a new key
      ENDSWITCH
    ENDIF
  ENDIF

  IF NOT ok.l THEN
        ; Place whatever error message or procedure you like to use here.
    SWITCH
      CASE error.n = 3 :     ; Possible Write Lock.
        TopMsg.u("Table locked by: "+ERRORUSER()+ ". Cannot verify key.","c",1000)
      OTHERWISE:
        TopMsg.u("A record with this key already exists.", "c", 1000)
    ENDSWITCH
  ENDIF
  IF requiredcheck.l <> false THEN
    RequiredCheck ON
  ENDIF
  MOVETO FIELD field.a ; Paradox3 sometimes jumps to another field or page.
  RETURN ok.l
ENDPROC
WRITELIB libname.a okkey35.l
RELEASE PROCS      okkey35.l


;-------------------------------------------------------------------------
; OkKey3.l()
; Purpose: Checks for key violation in coedit mode.
; Note: It is OkKey3 because it uses Paradox 3.0 and 3.01 commands.
; If you are using Paradox 3.5, this proc will automatically invoke the
; OkKey35.l() procedure which takes advantage of the newer environment.
;
; The proc returns a TRUE if the key is okay, and a FALSE if not.
;
; If the record is a Master record, with dependent detail records,
; and a key conflict exists, that is the state that will remain.
; It is safe, because subsequent calls to Okkey3.l will *NOT*
; overwrite any existing record because it does not use a LOCKRECORD
; to test for ability to post.
;
; If not in a Master record with active detail records, the proc
; deletes the record and then reinserts it, returning it to a
; "not yet posted" state.  This avoids a potential problem when
; a record is locked and the next attempt to lock or unlock the
; record results in "key already exists" situation. If this occurs,
; the next LOCKRECORD or UNLOCKRECORD will automatically post
; the offending record, overwriting the record with the valid
; key.  This is not usually what you want, so this proc keeps
; the record as not yet posted, to make it relatively harmless.
;
; A byproduct of the delete and reinsert technique is that UNDO
; becomes dangerous, as it would remove the recently inserted
; record.  This proc gets around this by always keeping a copy
; of the record (with COPYTOARRAY) being edited. When the user
; wants to "Undo" changes, it issues a COPYFROMARRAY to restore
; the record's original state.
;
; Note that this routine ONLY works with Paradox version 3, as it
; uses the RESYNCKEY command to pseudo-post the record without letting
; it fly away.
;
; This current incarnation also traps for the way Paradox 3.01 handles
; the KEYLOOKUP command in master tables (KeyLookup has been disabled, and a
; new ERRORMESSAGE has been instated).  The procedure is backwards compatable
; with Paradox 3.0.
;
;Examples l = OkKey3.l(TRUE)   ; Make sure REQUIREDCHECK is turned back on.
;         l = OkKey3.l(FALSE)  ; REQUIREDCHECK will remain off.
;         l = OkKey3.l("")     ; REQUIREDCHECK will be turned back ON.
;

PROC OkKey3.l(requiredcheck.l)   ; version 3.10
    ; requiredcheck.l  ; Determines if required fields should be
    ;                  ; reinstated when leaving the proc.
                       ; Should be TRUE if you want to make sure that
                       ; REQUIREDCHECK ON is in effect when exiting
                       ; Copyfromarray3.u, which is called by this proc.
                       ; If requiredcheck.l is any value other than FALSE,
                       ; REQUIREDCHECK will be set back to the normal
                       ; ON status.
    ; RETURNS TRUE if key is okay, FALSE if key already exists

  PRIVATE
  field.a,         ; Current field name.
  r,               ; Temporary array.
  a,               ; Window() message.
  ok.l,            ; Is record okay (no key conflict)?
  error.n,         ; Errorcode for keyviolation or write lock.
  proc.a           ; Name of procedure

  proc.a = "OkKey3.l"

    ; The next few lines allow you to specify an alternate okkey procedure,
    ; perhaps necessary under special circumstances, for example when an
    ; embedded table is also explicitly place on the workspace. To invoke
    ; an alternate key violation procedure, assign a variable called
    ; okkeyproc.a to the name of the alternate procedure prior to invoking
    ; the okkey3.l procedure. The alternate procedure must return a TRUE
    ; if the key is okay, or FALSE if there is a keyviolation.
    ;
    ; This can be handy for performance gains by making the key checking
    ; more specific to the particular needs of the table. Remember
    ; to release or reassign okkeyproc.a = "" when no longer needed.

  IF IsAssigned (okkeyproc.a) AND okkeyproc.a <> "" THEN
    EXECPROC okkeyproc.a    ; Do the special procedure and return its value.
    RETURN RETVAL
  ENDIF

  IF VERSION()>=3.5 THEN     ; Let's do the procedure that is sensitive to 3.5
    OkKey35.l(requiredcheck.l)
    RETURN RETVAL     ; Send the value back to the calling procedure
  ENDIF

  ; If no special keyfield proc has been assigned, we continue from this point
  IF NKeyFields(TABLE()) = 0 THEN
    ok.l = true      ; No key fields, cannot possibly be a key conflict.
  ELSE               ; There must be key fields.
        ; Now on to the keyviolation checking...
    RequiredCheck OFF     ; Helps deal with required valcheck and empty fields.
    CopyToArray r         ; For safekeeping.
    ImageRights           ; Make sure that we have rights to do this.
    DEL                   ; Attempt to remove the current record.
    SWITCH
      CASE SEARCH("Can't delete master", WINDOW()) = 1 : ; Must be a master table.
        field.a = FIELD()   ; Just in case RESYNCKEY manages to jump you to
                                    ; another page in a multi page form.
        resynckey           ; Shows a key conflict without unlocking-- records
                                    ; won't fly away.
        error.n = ERRORCODE()
        ok.l = error.n <> 53 AND error.n <> 3  ; Error.n 53 is set if key exists;
                                                       ; error.n 3 if a Write Lock is
                                                       ; on table.
        MOVETO FIELD field.a ; Paradox3 sometimes jumps to another field or page.
      OTHERWISE:             ; Detail record, or not a master with details.
        Ins                  ; Insert a blank, unposted record.
                                     ; Copy from the array to create a complete
                                     ; unposted record.
        CopyFromArray3.u("r", requiredcheck.l) ; checks for possible problems
        KeyLookup            ; Attempt toggle to possible offending key.
        error.n = ERRORCODE()
        a = ERRORMESSAGE()
        IF MATCH(a,"keylookup can not..") THEN  ; 3.01 doesn't allow KEYLOOKUP
                                                          ; in master.
          resynckey
          error.n = ERRORCODE()
          ok.l = error.n <> 53 AND error.n <> 3 ; error.n 53 is set if key exists;
                                                          ; error.n 3 if a Write Lock is
                                                          ; on table.
        ELSE
          ok.l = error.n = 56                  ; If it can't find lookup,
                                                          ; then key is okay.
        ENDIF
        IF ok.l THEN         ; Lookupkey not found.
          LockRecord         ; Record posts, and cursor remains with it.
        ELSE
          KeyLookup          ; Return to record values as operator entered.
        ENDIF
    ENDSWITCH
  ENDIF
  IF NOT ok.l THEN
        ; Place whatever error message or procedure you like to use here.
    SWITCH
      CASE error.n = 3 :     ; Possible Write Lock.
        TopMsg.u("Table locked by: "+ERRORUSER()+ ". Cannot verify key.","C",1000)
      OTHERWISE:
        TopMsg.u("A record with this key already exists.", "C", 1000)
    ENDSWITCH
  ENDIF
  IF requiredcheck.l <> false THEN
    RequiredCheck ON
  ENDIF
  RETURN ok.l
ENDPROC
WRITELIB libname.a OkKey3.l
RELEASE PROCS      OkKey3.l
;========================================================================

;========================================================================
; CopyFromArray3.u()
; Version 1.5. This revision supercedes all previous versions.
;
; This procedure simulates a COPYFROMARRAY that will work in
; all table images whether master, linked, or unlinked. It uses
; ErrorProc to specify a procedure to handle errors that would
; normally occur when using the COPYFROMARRAY command. For those
; who have used prior versions, the syntax is the same, but
; there are some new features of CopyFromArray3.u :

; 1) Faster than previous versions.
; 2) Works with all forms of linked embedded tables.
; 3) Works in all edit modes, not just CoEdit.
; 4) Lets Paradox version 3.0 place blank required fields
;    (this caused a script error in Paradox 3.0, but was fixed
;    in Paradox 3.01).
;
; 5) If an unanticipated error occurs, the OopsPrint.u()
;    proc user is presented with a
;    Cancel/Debug prompt (just as COPYFROMARRAY would).
; 6) In the event of an error caused by invalid data, a procedure called
;    InvalidArrayRecover.u() is invoked to document the error and try to
;    recover any potentially lost data.

PROC CopyFromArray3.u(arrayname.a, requiredcheck.l)
  ; arrayname.a      ; A string representing the name of the array to copy from.
  ; requiredcheck.l  ; If you set requiredcheck.l to FALSE, REQUIREDCHECK
  ;                    will remain OFF.
  ;                    If requiredcheck.l is any other value, REQUIREDCHECK
  ;                    will be turned back to its normal ON status.
  ; linkedkeys.n       Optional global variable, if set, tells CopyFromaArray3.u
  ;                    how many keys to avoid placing.  This is necessary only in
  ;                    one odd circumstance described below.
  PRIVATE
  fields.n,   ; Number of fields in table.
  n ,         ; Field counter.
  linked.n,   ; Number of linked keys specified by optional variable
                ; called linkedkeys.n   ; see notes for explanation
  ERRORPROC,  ; Name of error procedure.
  proc.a      ; Name of procedure

  proc.a = "CopyFromArray3.u"

    ; Requires you to define an error procedure to handle link and required
    ; field errors. We've tried to make the proc as small as possible because
    ; it is only called when an EXECUTE fails in the code below, and EXECUTEs
    ; totally disables procedure swapping.
  PROC copyerror.n()
    PRIVATE n   ; Used by GETCHAR() in case of error.
    SWITCH
              ; The first case statement traps for linked fields, empty master records
              ; and required fields.
      CASE (ERRORCODE() = 60 AND
        MATCH (ERRORMESSAGE(),"..linked fields in ..") OR
        MATCH (ERRORMESSAGE(),"..master record is blank..")) OR
        (ERRORCODE() = 23 AND MATCH(ERRORMESSAGE(),"..value must be provided..")):
        RETURN 1
      CASE ERRORCODE() = 23 and MATCH(ERRORMESSAGE(),"..not one of the possible value.."):
        ; Do the recovery for invalid data
        InvalidArrayRecover.u(arrayname.a)
        RETURN 2   ; The cancel debug prompt
      OTHERWISE: ; An unanticipated script error.
        ; This could occur, for example, if another user has locked the
        ; current record
         MESSAGE "CopyFromArray3.u has found Errorcode() =",
                  ERRORCODE(), "  ", ERRORMESSAGE()+"   Press any key."
                  n = GETCHAR()
        RETURN 2  ; Causes Cancel/Debug prompt to appear after key is pressed.
                  ; In Runtime, it would simply end script play and exit to DOS.
    ENDSWITCH
  ENDPROC


  RequiredCheck OFF           ; You might not even be allowed to issue
                                ; MENU otherwise.
  ERRORPROC = "CopyError.n"   ; Set name of error handler

  IF VERSION() >= 3.5 THEN
      ; Linked problems with COPYFROMARRAY were fixed in 3.5
      ; however the invalid data problems could still be present
    EXECUTE "COPYFROMARRAY " + arrayname.a
  ELSE
     ; let's trap for potential problems with PAradox 3.0 and 3.01
    fields.n = NFields(TABLE())

      ; The next few lines allow you to specify the number of linked keys on an
      ; embedded form.  Normally this is not necessary. However, there is a bug
      ; in Paradox3 whereby if a table is present as an embedded form and *also*
      ; is elsewhere on the workspace, attempting to assign a linked key field
      ; moves the cursor out of the embedded form and onto the tableview image.
      ; This also happens with the COPYFROMARRAY command.  So, if you have set up
      ; the workspace so that this becomes a problem, set a variable called
      ; linkedkeys.n to the number of key fields to ignore during the copyfromarray
      ; process.  Remember to release or reset the linkedkeys.n variable to 0 if
      ; it doesn't apply to the table.
    IF NOT IsAssigned (linkedkeys.n) THEN
      linked.n = 0
    ELSE
      linked.n = linkedkeys.n
    ENDIF

      ; Use the {Image} {Zoom} menu to find out field names and place the
      ; values into the record. Luckily, the menu choices presented are not
      ; sensitive to column rotations.

    Menu {Image} {Zoom} {Field}
      ; Now, make an attempt place all available fields. If no linked keys
      ; have been specified with the linkedkeys.n variable, the errorproc will
      ; handle linked fields and ignore them.

    ; Note that all fields will be filled, even if they are not different from
    ; the array element.  This is done because there might be some fill-in fields
    ; that are sensitive to the order in which the values are placed.  That is,
    ; Field 1 may fill in field 2, but field 2 may need to fill in field 3.  If
    ; we did not explicitly assign the fields when unchanged, the subsequent
    ; fill-ins would not work the same as the COPYFROMARRAY command.
    ; Granted, it's an obscure point, but one worth mentioning.

    FOR n FROM 1 TO fields.n         ; Step through the field names one at a time.
      IF n > linked.n THEN
        EXECUTE "[" + MENUCHOICE() + "]=" + arrayname.a + "[n+1]"
         ; For example, if the first field is [Cust#] the execute would prepare:
         ; [Cust#]=record.r[2]      ; The second element of the array.
      ENDIF
      Right                            ; Go on to the next field name.
    ENDFOR
    CTRLBREAK                          ; Back out of the zoom/field.
  ENDIF
  IF requiredcheck.l <> false THEN
    REQUIREDCHECK ON
  ENDIF
  RELEASE PROCS copyerror.n         ; Remove the error proc from memory.
ENDPROC
WRITELIB libname.a CopyFromArray3.u
RELEASE PROCS      CopyFromArray3.u

;----------------------------------------------------------------------
; InvalidArrayRecover.u()
;
; This proc is only invoked when there has been an invalid field assignment
; through a COPYFROMARRAY.  It is called by CopyFromarray3.u() whenever
; invalid data causes a script error while attempting to update a field.
; Existing data that does not pass the validity checks can be invalid due to
; several factors.
;  1) A LowValue or HighValue valcheck was changed after the data was entered,
;     causing the current data to be out of range.
;  2) The field has a TableLookup valcheck, and the current value no longer
;     exists in the lookup table.
;
; Obviously it is much better never to run into this case in the first place,
; but sometimes we come across data that was not entered through script control
; and the invalid data should not simply be ignored.
;
; This procedure attempts to recover from the error, and document all
; the values in the array.  Because of some quirks in Paradox, it is impossible
; to determine just where the cursor will end up in the event of the run error,
; so we cannot simply recover and continue.
;
; We will note the error in the ErrorLog.Sc (or create the script if necessary).
; It will document all the values in the offending array and the postion of
; the cursor at the time of the error (this is not necessarily the
; offending field).
; Next it will attempt to force the data back into the table by creating a
; temporary table and issuing an ADD command to override the valchecks.

Proc InvalidArrayRecover.u(arrayname.a)
 PRIVATE
  Errorproc,      ; This is an error procedure, so we keep ErrorProc private
  Oldtable.a,     ; Current table name
  TempTable.a,    ; Name of temporary table
  Recno.n,        ; Current record number
  Formview.l,     ; Are we in form view?
  Field.a,        ; Current field name
  Sysmode.a,      ; Sysmode at time of error
  zzzxzzzz.r      ; Values of current record

 ; First let's alert the user that there is a problem
 TopMsg.u("Error: There is a invalid data in an array",
          "Press any key and wait while a copy of the data is written to the ErrorLog.",
          1000)

 TopMsg.u("Writing contents of array" + arrayname.a ,
          "to ErrorLog.sc.  Please wait.","")

  Oldtable.a = TABLE()   ; Find the original table name
  Recno.n = RECNO()
  Formview.l = ISFORMVIEW()   ; Lets us put it back.
  Field.a = FIELD()
  Sysmode.a = SYSMODE()

 ; Let's create a holding array
 EXECUTE "size.n = ARRAYSIZE(" + arrayname.a +")"
 PRINT FILE "ErrorLog.sc"
 "\n**Error** Invalid Data while attempting CopyFromArray ", arrayname.a,
 "\n       Date: ", TODAY(),
 "\n       Time: ", TIME(),
 "\n       User: ", USERNAME(),
 "\n      Table: ", TABLE(),
 "\n      Recno: ", RECNO(),
 "\n      Field: ", FIELD(),
 "\n  ErrorCode: ", ERRORCODE(),
 "\n  ErrorMess: ", ERRORMESSAGE(),
 "\n\n  ARRAY ", arrayname.a, "[", size.n, "]\n"

 FOR n FROM 1 TO size.n
   ; get the value
   EXECUTE "v = " + arrayname.a +"[n]"
   PRINT FILE "ErrorLog.sc"
   "  ", arrayname.a, "[", n, "] = ", v, "\n"
 ENDFOR

 PRINT FILE "ErrorLog.sc"
 "\n  Current field values in ", TABLE(), ".db record at time of error:"

 COPYTOARRAY zzzxzzzz.r   ; The current record
 FOR n FROM 2 to ARRAYSIZE(zzzxzzzz.r)
   ; get the value
   PRINT FILE "ErrorLog.sc"
   "\n  Field #", n-1, " = ", zzzxzzzz.r[n]
 ENDFOR

 ; Now let's try a DO_IT! to get out of our editing mode
 TopMsg.u("Attempting to recover data and add it to the table",
          "Please wait","")
 DO_IT!
 IF Sysmode() <> "Main" THEN  ; the DO_IT was unsuccessful
   UNDO   ; Either deletes new record, or restores original unedited record
   DO_IT!
 ENDIF
 image.n = imageno()
 ; To get the invalid data back into the table we will create a temporary
 ; table like the current table.  Then copy the data into the new table, and
 ; finally ADD the new table to the old table, skirting the ValChecks.
 TempTable.a = PRIVDIR() + "ZZZXZZZZ"
 CREATE TempTable.a LIKE Oldtable.a
 CoEdit TempTable.a
 EXECUTE "Copyfromarray " + arrayname.a
 DO_IT!
 ADD TempTable.a OldTable.a
 ; Updates the existing table with record information
 If TABLE() = "Changed" THEN
  DELETE "Changed"
 ENDIF
 Delete TempTable.a
 Moveto image.n
 IF SEARCH(sysmode.a,"CoEdit")>0 THEN
   EXECUTE sysmode.a +"Key"    ; EditKey   or CoEditKey
 ENDIF
 IF formview.l THEN
   FORMKEY
 ENDIF
 MOVETO oldtable.a  ; Could have been a detail table
 TopMsg.u("All possible attempts have been made to recover the data",
          "Examine data when possible to verify recovery.  Press any key." , 0)
 TopMsg.u("Error conditon makes it impossible to continue program because the ",
          "data may have been affected.  Press any key for Debug prompt", 0)
 ; The procedure that calls this one is responsible for final action,
 ; which is normally presenting the Cancel/Debug prompt
ENDPROC
WRITELIB libname.a InvalidArrayRecover.u
RELEASE PROCS      InvalidArrayRecover.u

;-------------------------------------------------------------------------

; This routine can only be used when there is no linked forms involved.
; It is potentially faster than OkKey3.l() because it does less error
; checking .   See notes about Key Violation checking under OkKey3.l()
; The procedure also does not check for invalid data, so use only when
; appropriate.

PROC OkKey2.l()     ;assumes coedit mode and returns a true or false
  PRIVATE
  conflict.r, ; Temporary array
  proc.a      ; Name of procedure

  proc.a = "OkKey2.l"

  ECHO OFF
  CopyToArray conflict.r
  DEL
  Ins            ; need an unposted record to
  CopyFromArray conflict.r
  KeyLookup     ; attempt lookup toggle to possible offending key
  IF ERRORCODE()=56 THEN   ; lookupkey not found
    LockRecord
    RETURN true
  ENDIF
  KeyLookup  ; return to record values as operator entered
    ; place whatever error proc you use to alert your user here
  TopMsg.u("A record with this key is already on file.",
  "Press any key to continue",1000)
  RETURN false
ENDPROC

WRITELIB libname.a OkKey2.l
RELEASE PROCS      OkKey2.l


;-------------------------------------------------------------------------
; OopsPrint.u()
; We have found that this procedure saves time and eliminates a great
; deal of confusion.
;
; Paradox 2 and 3 allows for the setting of a special system variable
; called ERRORPROC. ERRORPROC contains the name of a procedure that was
; read in from a library that will execute when a run error occurrs.
; This variation of OopsPrint.u is used to document UNANTICIPATED errors.
;
; Appearances are deceptive. This procedure was developed to accurately
; document the Paradox environment when an unanticipated run error causes
; a script to fail. Usually there is a dramatic difference between what
; an operator thought happened and what really happened.
;
; For example, suppose your script was controlling several tables on
; the workspace in CoEdit mode. A MOVETO "tablename" was ignored
; because of an unexpected key violation (resulting in errors in
; subsequent field assignments). The odds are that the user would not
; be viewing the table where the trouble was occurring (instead
; perhaps, the last table before the PAL canvas was dropped).
;
; This procedure should be put into place after all known bugs have
; been found. When an unknown bug (or table corruption) is found,
; script play stops, and the complete status of the system is
; output to the printer (if online), and appended to a script called
; ERRORLOG.SC in the current directory. You might want to change the
; name of the script, or its location, to suit the application.
;
; When the client calls with a problem, we usually ask them to read
; the piece of paper that was just ejected from the printer.  We also
; ask them what they were doing just before the error occurred.  The
; ERRORLOG lets you see a history of errors, (many of which your clients
; will not admit to), and lets you get a feel of any pattern that is
; developing (whether or not it is script based, or errors in the
; integrity of the data or indexes).
;
; OopsPrint.u() gets all appropriate information about the system status,
; but must be read in at the beginning of the application. This allows
; Paradox to find it when necessary (even though it will be swapped out of
; memory later).
;
;  Example:    READLIB libname.a OopsPrint.u, working.u, beeper.u
;
; Alternately, OopsPrint.u() may be called from another errorproc specific
; to your application when the errorproc cannot deal with the error found.
; For example:
; PROC Oops.n()
;   PRIVATE Errorproc       ; So any errors in this proc won't be recursive.
;   SWITCH
;     CASE ERRORCODE() = 41:; "Not enough disk space to complete operation."
;        ;statements        ; Do something to free up some disk space.
;        RETURN 0           ; To retry the statement that caused the error.
;     OTHERWISE:
;        OopsPrint.u()      ; Print the errors.
;   ENDSWITCH
; ENDPROC
;
; You'll may notice that one thing that OopsPrint.u *DOES NOT* do is
; to issue a SAVEVARS ALL. There are several reasons. First, your code
; might make use of a SAVEVARS.SC, and we did not want to overwrite it.
; Second, there there was a bug involving SAVEVARS in all Paradox versions
; prior release 3.0. The bug mainfests itself by locking up the machine
; when the the SAVEVARS command attempts to write a line in SAVEVARS.SC
; that is longer than 132 characters.  This presents a problem with long
; alphanumeric variables (such as TkNegKey and TkPosKey used in the
; DataEntry Toolkit) as well as long text fields assigned with COPYTOARRAY
; statements. The bug was *almost* fixed in Paradox version 3, so you might
; want to include some code like:
;          IF VERSION() > = 3.0 THEN
;             SAVEVARS ALL
;          ENDIF
; It will still cause a problem if you are using long variable that contain
; many characters with ASCII values below 30.  This is of particular
; importance to those who use the Data Entry Toolkit with record level
; movement procedures.
;
; This procedure uses a lot of variables and functions, so it consumes about
; 9kb of memory.  If you are working within very limited memory, you
; might not even have enough memory to load this procedure in the event of a run
; error.  You might consider splitting the procedure into two parts, the first
; part could get all the variables, and the second part could record them to
; disk and printer.

PROC OopsPrint.u ()   ; error procedure
  PRIVATE
  ERRORPROC,              ; Private in case there is error in this proc. We
                            ; would not want the errorproc to call itself.
  sysmode.a,              ; System mode  (EDIT, MAIN, etc).
  drivespace.n,           ; Amount of free space on disk.
  date.d,                 ; Current date.
  time.a,                 ; Current time.
  version.n,              ; Paradox Version (even though accurate to only 1 decimal).
  errorcode.n,            ; Paradox error code.
  errormessage.a,         ; Paradox error message.
  erroruser.a,            ; Network user who caused the error (i.e. locked record).
  username.a,             ; Current network user.
  memleft.n,              ; Bytes of free memory for procs and variables.
  nimages.v,              ; # of images on the workspace.
  table.a,                ; Table name (if applicable).
  imagetype.a,            ; Image type (if applicable).
  field.a,                ; Field name (if applicable).
  nrecords.v,             ; Number of table records (if applicable).
  nimagerecords.v,        ; Number of image records (if applicable).
  form.a,                 ; Form name (if applicable)
  recno.v,                ; Current record # (if applicable).
  fieldvalue.v,           ; Current field value (if applicable).
  menuchoice.a,           ; Current Paradox menuchoice or "Error"
  isvalid.a,              ; Is the currecnt field valid?
  runtime.a,              ; Is this paradox runtime? (converted to string).
  printerstatus.a,        ; Is printer online and ready?.
  directory.a,            ; Current directory.
  privdir.a,              ; Private directory.
  isshared.l,             ; Is current table shared?.
  n,                      ; Used to remove keys from keyboard buffer.
  procname.a              ; Used to display Proc.a variable (if assigned)

  BEEP
  SLEEP 200  BEEP
  SLEEP 200  BEEP
  MESSAGE "Error has occurred, please wait while it is logged"
  WHILE CHARWAITING()   ; Clear keyboard buffer, so user does not inadvertantly.
                            ; Press Cancel if the {Cancel}/{Debug} prompt is displayed.
    n = GETCHAR()
  ENDWHILE
  directory.a = DIRECTORY()
  privdir.a = PrivDir()
  sysmode.a = SYSMODE()
  drivespace.n = DriveSpace (SUBSTR (directory.a, 1, 1) )
        ;-- First character of current directory is drive
  date.d = TODAY()
  time.a = TIME()
  version.n = VERSION()
  errorcode.n = ERRORCODE()
  errormessage.a = ERRORMESSAGE()
  erroruser.a = ERRORUSER()
  username.a = UserName()
  memleft.n = MEMLEFT()
  runtime.a = FORMAT("LY",IsRunTime())
  printerstatus.a = FORMAT("LO",PrinterStatus())  ; "On" or "Off"
  IF NImages()<>0 THEN       ; Occurred on image on workspace
    nimages.v = NImages()
    table.a = TABLE()
    imagetype.a = ImageType()
    field.a = FIELD()
    SWITCH
      CASE imagetype.a = "Query":
        IF CheckMarkStatus()<>"" THEN  ; Place checkmark if appropriate
          fieldvalue.v = CheckMarkStatus()+" "+[]
        ELSE
          fieldvalue.v = []
        ENDIF
      CASE nimagerecords() > 0 : ; Cannot get value from empty table
        fieldvalue.v = FieldStr()
      OTHERWISE:      ; Must be an empty table
        fieldvalue.v = ""
    ENDSWITCH
    isshared.l = IsShared(table.a)
    nrecords.v = NRecords(TABLE())
    nimagerecords.v = nimagerecords()
    IF imagetype.a ="Display" AND SEARCH(SYSMODE(),"MainCoEditDataEntry")>0 THEN
      form.a = form()
      recno.v = RECNO()
    ELSE ; Could be in restructure subsystem where Formview is inappropriate.
      form.a = "N/A"
      recno.v = RECNO()
    ENDIF
  ELSE                ; Not in an image.
    nimages.v = "N/A"
    table.a = "N/A"
    imagetype.a = "N/A"
    isshared.l = "N/A"
    field.a = "N/A"
    nrecords.v = "N/A"
    nimagerecords.v = "N/A"
    form.a ="N/A"
    recno.v = "N/A"
    fieldvalue.v = "N/A"
  ENDIF
  menuchoice.a = MENUCHOICE()  ; The current Paradox menu choice or "Error"
    ; Now we're checking to see if a variable called Proc.a is assigned.  Many
    ; programmers set this variable in all their procedures in order to have the
    ; procedure name available in the event of a run error.
  IF IsAssigned(proc.a) THEN
    procname.a = proc.a
  ELSE
    procname.a = ""
  ENDIF

  PRINT FILE "ERRORLOG.SC"    ; In the current directory (appends if one exists).
  "\n\n" ,                    ; "\n" is linefeed  "\f"  is formfeed.
  "           Date: ", date.d, "\n",
  "           Time: ", time.a, "\n",
  "        Version: ", version.n, "\n",
  "      Errorcode: ", errorcode.n, "\n",
  "   Errormessage: ", errormessage.a, "\n",
  "      Erroruser: ", erroruser.a, "\n",
  "           User: ", username.a, "\n",
  "         Proc.a: ", procname.a, "\n",
  "    Memory left: ", memleft.n, "\n",
  "        Sysmode: ", sysmode.a, "\n",
  "        Nimages: ", nimages.v, "\n",
  "          Table: ", table.a, "\n",
  "         Shared: ", isshared.l, "\n",
  "     Image Type: ", imagetype.a, "\n" ,
  "     Field name: ", field.a, "\n",
  "       Nrecords: ", nrecords.v, "\n",
  "  NImageRecords: ", nimagerecords.v, "\n",
  "           Form: ", form.a, "\n",
  "       Record #: ", recno.v, "\n",
  "    Field value: ", fieldvalue.v, "\n",
  "     MenuChoice: ", menuchoice.a, "\n",
  "        Runtime: ", runtime.a, "\n",
  "     Drivespace: ", drivespace.n, "\n",
  "      Directory: ", directory.a, "\n",
  "        Privdir: ", privdir.a, "\n",
  " Printer Status: ", printerstatus.a ,"\n\n"

  IF PrinterStatus() THEN   ; Print out status if printer is available.
    open PRINTER
        ; Keeps [Crtl Z] from printing as right arrows on each line,
        ; if you have Paradox2 on a HP LaserJet II in IBM character set mode.
    PRINT
    "\f",    ; a form feed
    "** Error occured in program **", "\n\n",
    "\n\n" ,
    "           Date: ", date.d, "\n",
    "           Time: ", time.a, "\n",
    "        Version: ", version.n, "\n",
    "      Errorcode: ", errorcode.n, "\n",
    "   Errormessage: ", errormessage.a, "\n",
    "      Erroruser: ", erroruser.a, "\n",
    "           User: ", username.a, "\n",
    "         Proc.a: ", procname.a, "\n",
    "    Memory left: ", memleft.n, "\n",
    "        Sysmode: ", sysmode.a, "\n",
    "        Nimages: ", nimages.v, "\n",
    "          Table: ", table.a, "\n",
    "         Shared: ", isshared.l, "\n",
    "     Image Type: ", imagetype.a, "\n" ,
    "     Field name: ", field.a, "\n",
    "       Nrecords: ", nrecords.v, "\n",
    "  NImageRecords: ", nimagerecords.v, "\n",
    "           Form: ", form.a, "\n",
    "       Record #: ", recno.v, "\n",
    "    Field value: ", fieldvalue.v, "\n",
    "     MenuChoice: ", menuchoice.a, "\n",
    "        Runtime: ", runtime.a, "\n",
    "     Drivespace: ", drivespace.n, "\n",
    "      Directory: ", directory.a, "\n",
    "        Privdir: ", privdir.a, "\n",
    " Printer Status: ", printerstatus.a ,
    "\f"   ; a form feed
    close PRINTER
  ENDIF

ENDPROC
WRITELIB libname.a OopsPrint.u
RELEASE PROCS      OopsPrint.u

